from typing import Optional
import mimetypes
import os.path
import pickle
import re
import sys
import time
from tempfile import NamedTemporaryFile
import warnings

import sacred.optional as opt
from sacred.commandline_options import cli_option
from sacred.dependencies import get_digest
from sacred.observers.base import RunObserver
from sacred.observers.queue import QueueObserver
from sacred.serializer import flatten
from sacred.utils import ObserverError, PathType

DEFAULT_MONGO_PRIORITY = 30

# This ensures consistent mimetype detection across platforms.
mimetypes.init(files=[])


def force_valid_bson_key(key):
    key = str(key)
    if key.startswith("$"):
        key = "@" + key[1:]
    key = key.replace(".", ",")
    return key


def force_bson_encodeable(obj):
    import bson

    if isinstance(obj, dict):
        try:
            bson.BSON.encode(obj, check_keys=True)
            return obj
        except bson.InvalidDocument:
            return {
                force_valid_bson_key(k): force_bson_encodeable(v)
                for k, v in obj.items()
            }

    elif opt.has_numpy and isinstance(obj, opt.np.ndarray):
        return obj
    else:
        try:
            bson.BSON.encode({"dict_just_for_testing": obj})
            return obj
        except bson.InvalidDocument:
            return str(obj)


class MongoObserver(RunObserver):
    COLLECTION_NAME_BLACKLIST = {
        "fs.files",
        "fs.chunks",
        "_properties",
        "system.indexes",
        "search_space",
        "search_spaces",
    }
    VERSION = "MongoObserver-0.7.0"

    @classmethod
    def create(cls, *args, **kwargs):
        warnings.warn(
            "MongoObserver.create(...) is deprecated. "
            "Please use MongoObserver(...) instead.",
            DeprecationWarning,
        )
        return cls(*args, **kwargs)

    def __init__(
        self,
        url: Optional[str] = None,
        db_name: str = "sacred",
        collection: str = "runs",
        overwrite: Optional[bool] = None,
        priority: int = DEFAULT_MONGO_PRIORITY,
        client: Optional["pymongo.MongoClient"] = None,
        failure_dir: Optional[PathType] = None,
        **kwargs
    ):
        """Initializer for MongoObserver.

        Parameters
        ----------
        url
            Mongo URI to connect to.
        db_name
            Database to connect to.
        collection
            Collection to write the runs to. (default: "runs").
        overwrite
            _id of a run that should be overwritten.
        priority
            (default 30)
        client
            Client to connect to. Do not use client and URL together.
        failure_dir
            Directory to save the run of a failed observer to.
        """
        import pymongo
        import gridfs

        if client is not None:
            if not isinstance(client, pymongo.MongoClient):
                raise ValueError(
                    "client needs to be a pymongo.MongoClient, "
                    "but is {} instead".format(type(client))
                )
            if url is not None:
                raise ValueError("Cannot pass both a client and a url.")
        else:
            client = pymongo.MongoClient(url, **kwargs)
        database = client[db_name]
        if collection in MongoObserver.COLLECTION_NAME_BLACKLIST:
            raise KeyError(
                'Collection name "{}" is reserved. '
                "Please use a different one.".format(collection)
            )
        runs_collection = database[collection]
        metrics_collection = database["metrics"]
        fs = gridfs.GridFS(database)
        self.initialize(
            runs_collection,
            fs,
            overwrite=overwrite,
            metrics_collection=metrics_collection,
            failure_dir=failure_dir,
            priority=priority,
        )

    def initialize(
        self,
        runs_collection,
        fs,
        overwrite=None,
        metrics_collection=None,
        failure_dir=None,
        priority=DEFAULT_MONGO_PRIORITY,
    ):
        self.runs = runs_collection
        self.metrics = metrics_collection
        self.fs = fs
        if isinstance(overwrite, (int, str)):
            overwrite = int(overwrite)
            run = self.runs.find_one({"_id": overwrite})
            if run is None:
                raise RuntimeError(
                    "Couldn't find run to overwrite with " "_id='{}'".format(overwrite)
                )
            else:
                overwrite = run
        self.overwrite = overwrite
        self.run_entry = None
        self.priority = priority
        self.failure_dir = failure_dir

    @classmethod
    def create_from(cls, *args, **kwargs):
        self = cls.__new__(cls)  # skip __init__ call
        self.initialize(*args, **kwargs)
        return self

    def queued_event(
        self, ex_info, command, host_info, queue_time, config, meta_info, _id
    ):
        if self.overwrite is not None:
            raise RuntimeError("Can't overwrite with QUEUED run.")
        self.run_entry = {
            "experiment": dict(ex_info),
            "command": command,
            "host": dict(host_info),
            "config": flatten(config),
            "meta": meta_info,
            "status": "QUEUED",
        }
        # set ID if given
        if _id is not None:
            self.run_entry["_id"] = _id
        # save sources
        self.run_entry["experiment"]["sources"] = self.save_sources(ex_info)
        self.insert()
        return self.run_entry["_id"]

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):
        if self.overwrite is None:
            self.run_entry = {"_id": _id}
        else:
            if self.run_entry is not None:
                raise RuntimeError("Cannot overwrite more than once!")
            # TODO sanity checks
            self.run_entry = self.overwrite

        self.run_entry.update(
            {
                "experiment": dict(ex_info),
                "format": self.VERSION,
                "command": command,
                "host": dict(host_info),
                "start_time": start_time,
                "config": flatten(config),
                "meta": meta_info,
                "status": "RUNNING",
                "resources": [],
                "artifacts": [],
                "captured_out": "",
                "info": {},
                "heartbeat": None,
            }
        )

        # save sources
        self.run_entry["experiment"]["sources"] = self.save_sources(ex_info)
        self.insert()
        return self.run_entry["_id"]

    def heartbeat_event(self, info, captured_out, beat_time, result):
        self.run_entry["info"] = flatten(info)
        self.run_entry["captured_out"] = captured_out
        self.run_entry["heartbeat"] = beat_time
        self.run_entry["result"] = flatten(result)
        self.save()

    def completed_event(self, stop_time, result):
        self.run_entry["stop_time"] = stop_time
        self.run_entry["result"] = flatten(result)
        self.run_entry["status"] = "COMPLETED"
        self.final_save(attempts=10)

    def interrupted_event(self, interrupt_time, status):
        self.run_entry["stop_time"] = interrupt_time
        self.run_entry["status"] = status
        self.final_save(attempts=3)

    def failed_event(self, fail_time, fail_trace):
        self.run_entry["stop_time"] = fail_time
        self.run_entry["status"] = "FAILED"
        self.run_entry["fail_trace"] = fail_trace
        self.final_save(attempts=1)

    def resource_event(self, filename):
        if self.fs.exists(filename=filename):
            md5hash = get_digest(filename)
            if self.fs.exists(filename=filename, md5=md5hash):
                resource = (filename, md5hash)
                if resource not in self.run_entry["resources"]:
                    self.run_entry["resources"].append(resource)
                    self.save()
                return
        with open(filename, "rb") as f:
            file_id = self.fs.put(f, filename=filename)
        md5hash = self.fs.get(file_id).md5
        self.run_entry["resources"].append((filename, md5hash))
        self.save()

    def artifact_event(self, name, filename, metadata=None, content_type=None):
        with open(filename, "rb") as f:
            run_id = self.run_entry["_id"]
            db_filename = "artifact://{}/{}/{}".format(self.runs.name, run_id, name)
            if content_type is None:
                content_type = self._try_to_detect_content_type(filename)

            file_id = self.fs.put(
                f, filename=db_filename, metadata=metadata, content_type=content_type
            )

        self.run_entry["artifacts"].append({"name": name, "file_id": file_id})
        self.save()

    @staticmethod
    def _try_to_detect_content_type(filename):
        mime_type, _ = mimetypes.guess_type(filename)
        if mime_type is not None:
            print(
                "Added {} as content-type of artifact {}.".format(mime_type, filename)
            )
        else:
            print(
                "Failed to detect content-type automatically for "
                "artifact {}.".format(filename)
            )
        return mime_type

    def log_metrics(self, metrics_by_name, info):
        """Store new measurements to the database.

        Take measurements and store them into
        the metrics collection in the database.
        Additionally, reference the metrics
        in the info["metrics"] dictionary.
        """
        if self.metrics is None:
            # If, for whatever reason, the metrics collection has not been set
            # do not try to save anything there.
            return
        for key in metrics_by_name:
            query = {"run_id": self.run_entry["_id"], "name": key}
            push = {
                "steps": {"$each": metrics_by_name[key]["steps"]},
                "values": {"$each": metrics_by_name[key]["values"]},
                "timestamps": {"$each": metrics_by_name[key]["timestamps"]},
            }
            update = {"$push": push}
            result = self.metrics.update_one(query, update, upsert=True)
            if result.upserted_id is not None:
                # This is the first time we are storing this metric
                info.setdefault("metrics", []).append(
                    {"name": key, "id": str(result.upserted_id)}
                )

    def insert(self):
        import pymongo.errors

        if self.overwrite:
            return self.save()

        autoinc_key = self.run_entry.get("_id") is None
        while True:
            if autoinc_key:
                c = self.runs.find({}, {"_id": 1})
                c = c.sort("_id", pymongo.DESCENDING).limit(1)
                self.run_entry["_id"] = c.next()["_id"] + 1 if c.count() else 1
            try:
                self.runs.insert_one(self.run_entry)
                return
            except pymongo.errors.InvalidDocument as e:
                raise ObserverError(
                    "Run contained an unserializable entry."
                    "(most likely in the info)\n{}".format(e)
                )
            except pymongo.errors.DuplicateKeyError:
                if not autoinc_key:
                    raise

    def save(self):
        import pymongo.errors

        try:
            self.runs.update_one(
                {"_id": self.run_entry["_id"]}, {"$set": self.run_entry}
            )
        except pymongo.errors.AutoReconnect:
            pass  # just wait for the next save
        except pymongo.errors.InvalidDocument:
            raise ObserverError(
                "Run contained an unserializable entry." "(most likely in the info)"
            )

    def final_save(self, attempts):
        import pymongo.errors

        for i in range(attempts):
            try:
                self.runs.update_one(
                    {"_id": self.run_entry["_id"]},
                    {"$set": self.run_entry},
                    upsert=True,
                )
                return
            except pymongo.errors.AutoReconnect:
                if i < attempts - 1:
                    time.sleep(1)
            except pymongo.errors.ConnectionFailure:
                pass
            except pymongo.errors.InvalidDocument:
                self.run_entry = force_bson_encodeable(self.run_entry)
                print(
                    "Warning: Some of the entries of the run were not "
                    "BSON-serializable!\n They have been altered such that "
                    "they can be stored, but you should fix your experiment!"
                    "Most likely it is either the 'info' or the 'result'.",
                    file=sys.stderr,
                )

        os.makedirs(self.failure_dir, exist_ok=True)
        with NamedTemporaryFile(
            suffix=".pickle",
            delete=False,
            prefix="sacred_mongo_fail_{}_".format(self.run_entry["_id"]),
            dir=self.failure_dir,
        ) as f:
            pickle.dump(self.run_entry, f)
            print(
                "Warning: saving to MongoDB failed! "
                "Stored experiment entry in '{}'".format(f.name),
                file=sys.stderr,
            )

    def save_sources(self, ex_info):
        base_dir = ex_info["base_dir"]
        source_info = []
        for source_name, md5 in ex_info["sources"]:
            abs_path = os.path.join(base_dir, source_name)
            file = self.fs.find_one({"filename": abs_path, "md5": md5})
            if file:
                _id = file._id
            else:
                with open(abs_path, "rb") as f:
                    _id = self.fs.put(f, filename=abs_path)
            source_info.append([source_name, _id])
        return source_info

    def __eq__(self, other):
        if isinstance(other, MongoObserver):
            return self.runs == other.runs
        return False


@cli_option("-m", "--mongo_db")
def mongo_db_option(args, run):
    """Add a MongoDB Observer to the experiment.

    The argument value is the database specification.
    Should be in the form:

    `[host:port:]db_name[.collection[:id]][!priority]`
    """
    kwargs = parse_mongo_db_arg(args)
    mongo = MongoObserver(**kwargs)
    run.observers.append(mongo)


def get_pattern():
    run_id_pattern = r"(?P<overwrite>\d{1,12})"
    port1_pattern = r"(?P<port1>\d{1,5})"
    port2_pattern = r"(?P<port2>\d{1,5})"
    priority_pattern = r"(?P<priority>-?\d+)?"
    db_name_pattern = r"(?P<db_name>[_A-Za-z]" r"[0-9A-Za-z#%&'()+\-;=@\[\]^_{}]{0,63})"
    coll_name_pattern = (
        r"(?P<collection>[_A-Za-z]" r"[0-9A-Za-z#%&'()+\-;=@\[\]^_{}]{0,63})"
    )
    hostname1_pattern = (
        r"(?P<host1>"
        r"[0-9A-Za-z](?:(?:[0-9A-Za-z]|-){0,61}[0-9A-Za-z])?"
        r"(?:\.[0-9A-Za-z](?:(?:[0-9A-Za-z]|-){0,61}"
        r"[0-9A-Za-z])?)*)"
    )
    hostname2_pattern = (
        r"(?P<host2>"
        r"[0-9A-Za-z](?:(?:[0-9A-Za-z]|-){0,61}[0-9A-Za-z])?"
        r"(?:\.[0-9A-Za-z](?:(?:[0-9A-Za-z]|-){0,61}"
        r"[0-9A-Za-z])?)*)"
    )

    host_only = r"^(?:{host}:{port})$".format(
        host=hostname1_pattern, port=port1_pattern
    )
    full = (
        r"^(?:{host}:{port}:)?{db}(?:\.{collection}(?::{rid})?)?"
        r"(?:!{priority})?$".format(
            host=hostname2_pattern,
            port=port2_pattern,
            db=db_name_pattern,
            collection=coll_name_pattern,
            rid=run_id_pattern,
            priority=priority_pattern,
        )
    )

    return r"{host_only}|{full}".format(host_only=host_only, full=full)


def parse_mongo_db_arg(mongo_db):
    g = re.match(get_pattern(), mongo_db).groupdict()
    if g is None:
        raise ValueError(
            'mongo_db argument must have the form "db_name" '
            'or "host:port[:db_name]" but was {}'.format(mongo_db)
        )

    kwargs = {}
    if g["host1"]:
        kwargs["url"] = "{}:{}".format(g["host1"], g["port1"])
    elif g["host2"]:
        kwargs["url"] = "{}:{}".format(g["host2"], g["port2"])

    if g["priority"] is not None:
        kwargs["priority"] = int(g["priority"])

    for p in ["db_name", "collection", "overwrite"]:
        if g[p] is not None:
            kwargs[p] = g[p]

    return kwargs


class QueueCompatibleMongoObserver(MongoObserver):
    def log_metrics(self, metric_name, metrics_values, info):
        """Store new measurements to the database.

        Take measurements and store them into
        the metrics collection in the database.
        Additionally, reference the metrics
        in the info["metrics"] dictionary.
        """
        if self.metrics is None:
            # If, for whatever reason, the metrics collection has not been set
            # do not try to save anything there.
            return
        query = {"run_id": self.run_entry["_id"], "name": metric_name}
        push = {
            "steps": {"$each": metrics_values["steps"]},
            "values": {"$each": metrics_values["values"]},
            "timestamps": {"$each": metrics_values["timestamps"]},
        }
        update = {"$push": push}
        result = self.metrics.update_one(query, update, upsert=True)
        if result.upserted_id is not None:
            # This is the first time we are storing this metric
            info.setdefault("metrics", []).append(
                {"name": metric_name, "id": str(result.upserted_id)}
            )

    def save(self):
        import pymongo

        try:
            self.runs.update_one(
                {"_id": self.run_entry["_id"]}, {"$set": self.run_entry}
            )
        except pymongo.errors.InvalidDocument:
            raise ObserverError(
                "Run contained an unserializable entry." "(most likely in the info)"
            )

    def final_save(self, attempts):
        import pymongo

        try:
            self.runs.update_one(
                {"_id": self.run_entry["_id"]}, {"$set": self.run_entry}, upsert=True
            )
            return

        except pymongo.errors.InvalidDocument:
            self.run_entry = force_bson_encodeable(self.run_entry)
            print(
                "Warning: Some of the entries of the run were not "
                "BSON-serializable!\n They have been altered such that "
                "they can be stored, but you should fix your experiment!"
                "Most likely it is either the 'info' or the 'result'.",
                file=sys.stderr,
            )

            with NamedTemporaryFile(
                suffix=".pickle", delete=False, prefix="sacred_mongo_fail_"
            ) as f:
                pickle.dump(self.run_entry, f)
                print(
                    "Warning: saving to MongoDB failed! "
                    "Stored experiment entry in '{}'".format(f.name),
                    file=sys.stderr,
                )

        raise ObserverError("Warning: saving to MongoDB failed!")


class QueuedMongoObserver(QueueObserver):
    """MongoObserver that uses a fault-tolerant background process."""

    @classmethod
    def create(cls, *args, **kwargs):
        warnings.warn(
            "QueuedMongoObserver.create(...) is deprecated. "
            "Please use QueuedMongoObserver(...) instead.",
            DeprecationWarning,
        )
        return cls(*args, **kwargs)

    def __init__(
        self,
        interval: int = 20,
        retry_interval: int = 10,
        url: Optional[str] = None,
        db_name: str = "sacred",
        collection: str = "runs",
        overwrite: Optional[bool] = None,
        priority: int = DEFAULT_MONGO_PRIORITY,
        client: Optional["pymongo.MongoClient"] = None,
        **kwargs
    ):
        """Initializer for MongoObserver.

        Parameters
        ----------
        interval
            The interval in seconds at which the background thread is woken up to process new events.
        retry_interval
            The interval in seconds to wait if an event failed to be processed.
        url
            Mongo URI to connect to.
        db_name
            Database to connect to.
        collection
            Collection to write the runs to. (default: "runs").
        overwrite
            _id of a run that should be overwritten.
        priority
            (default 30)
        client
            Client to connect to. Do not use client and URL together.
        failure_dir
            Directory to save the run of a failed observer to.
        """
        super().__init__(
            QueueCompatibleMongoObserver(
                url=url,
                db_name=db_name,
                collection=collection,
                overwrite=overwrite,
                priority=priority,
                client=client,
                **kwargs,
            ),
            interval=interval,
            retry_interval=retry_interval,
        )



import datetime as dt
import json
import os
from io import BufferedReader, FileIO

from hashfs import HashFS
from tinydb import TinyDB
from tinydb_serialization import Serializer, SerializationMiddleware

import sacred.optional as opt

# Set data type values for abstract properties in Serializers
series_type = opt.pandas.Series if opt.has_pandas else None
dataframe_type = opt.pandas.DataFrame if opt.has_pandas else None
ndarray_type = opt.np.ndarray if opt.has_numpy else None


class BufferedReaderWrapper(BufferedReader):
    """Custom wrapper to allow for copying of file handle.

    tinydb_serialisation currently does a deepcopy on all the content of the
    dictionary before serialisation. By default, file handles are not
    copiable so this wrapper is necessary to create a duplicate of the
    file handle passes in.

    Note that the file passed in will therefor remain open as the copy is the
    one that gets closed.
    """

    def __init__(self, f_obj):
        f_obj = FileIO(f_obj.name)
        super().__init__(f_obj)

    def __copy__(self):
        f = open(self.name, self.mode)
        return BufferedReaderWrapper(f)

    def __deepcopy__(self, memo):
        f = open(self.name, self.mode)
        return BufferedReaderWrapper(f)


class DateTimeSerializer(Serializer):
    OBJ_CLASS = dt.datetime  # The class this serializer handles

    def encode(self, obj):
        return obj.strftime("%Y-%m-%dT%H:%M:%S.%f")

    def decode(self, s):
        return dt.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f")


class NdArraySerializer(Serializer):
    OBJ_CLASS = ndarray_type

    def encode(self, obj):
        return json.dumps(obj.tolist(), check_circular=True)

    def decode(self, s):
        return opt.np.array(json.loads(s))


class DataFrameSerializer(Serializer):
    OBJ_CLASS = dataframe_type

    def encode(self, obj):
        return obj.to_json()

    def decode(self, s):
        return opt.pandas.read_json(s)


class SeriesSerializer(Serializer):
    OBJ_CLASS = series_type

    def encode(self, obj):
        return obj.to_json()

    def decode(self, s):
        return opt.pandas.read_json(s, typ="series")


class FileSerializer(Serializer):
    OBJ_CLASS = BufferedReaderWrapper

    def __init__(self, fs):
        self.fs = fs

    def encode(self, obj):
        address = self.fs.put(obj)
        return json.dumps(address.id)

    def decode(self, s):
        id_ = json.loads(s)
        file_reader = self.fs.open(id_)
        file_reader = BufferedReaderWrapper(file_reader)
        file_reader.hash = id_
        return file_reader


def get_db_file_manager(root_dir):
    fs = HashFS(os.path.join(root_dir, "hashfs"), depth=3, width=2, algorithm="md5")

    # Setup Serialisation object for non list/dict objects
    serialization_store = SerializationMiddleware()
    serialization_store.register_serializer(DateTimeSerializer(), "TinyDate")
    serialization_store.register_serializer(FileSerializer(fs), "TinyFile")

    if opt.has_numpy:
        serialization_store.register_serializer(NdArraySerializer(), "TinyArray")
    if opt.has_pandas:
        serialization_store.register_serializer(DataFrameSerializer(), "TinyDataFrame")
        serialization_store.register_serializer(SeriesSerializer(), "TinySeries")

    db = TinyDB(os.path.join(root_dir, "metadata.json"), storage=serialization_store)
    return db, fs



#!/usr/bin/env python
# coding=utf-8

from sacred.observers.base import RunObserver, td_format
from sacred.config.config_files import load_config_file
import logging


DEFAULT_TELEGRAM_PRIORITY = 10


class TelegramObserver(RunObserver):
    """Sends a message to Telegram upon completion/failing of an experiment."""

    @staticmethod
    def get_proxy_request(telegram_config):
        from telegram.utils.request import Request

        if telegram_config["proxy_url"].startswith("socks5"):
            urllib3_proxy_kwargs = dict()
            for key in ["username", "password"]:
                if key in telegram_config:
                    urllib3_proxy_kwargs[key] = telegram_config[key]
            return Request(
                proxy_url=telegram_config["proxy_url"],
                urllib3_proxy_kwargs=urllib3_proxy_kwargs,
            )
        elif telegram_config["proxy_url"].startswith("http"):
            cred_string = ""
            if "username" in telegram_config:
                cred_string += telegram_config["username"]
            if "password" in telegram_config:
                cred_string += ":" + telegram_config["password"]
            if len(cred_string) > 0:
                domain = telegram_config["proxy_url"].split("/")[-1].split("@")[-1]
                cred_string += "@"
                proxy_url = "http://{}{}".format(cred_string, domain)
                return Request(proxy_url=proxy_url)
            else:
                return Request(proxy_url=telegram_config["proxy_url"])
        else:
            raise Exception(
                "Proxy URL should be in format "
                "PROTOCOL://PROXY_HOST[:PROXY_PORT].\n"
                "HTTP and Socks5 are supported."
            )

    @classmethod
    def from_config(cls, filename):
        """
        Create a TelegramObserver from a given configuration file.

        The file can be in any format supported by Sacred
        (.json, .pickle, [.yaml]).
        It has to specify a ``token`` and a ``chat_id`` and can optionally set
        ``silent_completion``, ``started_text``, ``completed_text``,
        ``interrupted_text``, and ``failed_text``.
        """
        import telegram

        d = load_config_file(filename)
        request = cls.get_proxy_request(d) if "proxy_url" in d else None

        if "token" in d and "chat_id" in d:
            bot = telegram.Bot(d["token"], request=request)
            obs = cls(bot, **d)
        else:
            raise ValueError(
                "Telegram configuration file must contain "
                "entries for 'token' and 'chat_id'!"
            )
        for k in ["started_text", "completed_text", "interrupted_text", "failed_text"]:
            if k in d:
                setattr(obs, k, d[k])
        return obs

    def __init__(
        self,
        bot,
        chat_id,
        silent_completion=False,
        priority=DEFAULT_TELEGRAM_PRIORITY,
        **kwargs
    ):
        self.silent_completion = silent_completion
        self.chat_id = chat_id
        self.bot = bot

        self.started_text = (
            "♻ *{experiment[name]}* "
            "started at _{start_time}_ "
            "on host `{host_info[hostname]}`"
        )
        self.completed_text = (
            "✅ *{experiment[name]}* "
            "completed after _{elapsed_time}_ "
            "with result=`{result}`"
        )
        self.interrupted_text = (
            "⚠ *{experiment[name]}* " "interrupted after _{elapsed_time}_"
        )
        self.failed_text = (
            "❌ *{experiment[name]}* failed after "
            "_{elapsed_time}_ with `{error}`\n\n"
            "Backtrace:\n```{backtrace}```"
        )
        self.run = None
        self.priority = priority

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):
        import telegram

        self.run = {
            "_id": _id,
            "config": config,
            "start_time": start_time,
            "experiment": ex_info,
            "command": command,
            "host_info": host_info,
        }
        if self.started_text is None:
            return
        try:
            self.bot.send_message(
                chat_id=self.chat_id,
                text=self.get_started_text(),
                disable_notification=True,
                parse_mode=telegram.ParseMode.MARKDOWN,
            )
        except Exception as e:
            log = logging.getLogger("telegram-observer")
            log.warning("failed to send start_event message via telegram.", exc_info=e)

    def get_started_text(self):
        return self.started_text.format(**self.run)

    def get_completed_text(self):
        return self.completed_text.format(**self.run)

    def get_interrupted_text(self):
        return self.interrupted_text.format(**self.run)

    def get_failed_text(self):
        return self.failed_text.format(
            backtrace="".join(self.run["fail_trace"]), **self.run
        )

    def completed_event(self, stop_time, result):
        import telegram

        if self.completed_text is None:
            return

        self.run["result"] = result
        self.run["stop_time"] = stop_time
        self.run["elapsed_time"] = td_format(stop_time - self.run["start_time"])

        try:
            self.bot.send_message(
                chat_id=self.chat_id,
                text=self.get_completed_text(),
                disable_notification=self.silent_completion,
                parse_mode=telegram.ParseMode.MARKDOWN,
            )
        except Exception as e:
            log = logging.getLogger("telegram-observer")
            log.warning(
                "failed to send completed_event message via telegram.", exc_info=e
            )

    def interrupted_event(self, interrupt_time, status):
        import telegram

        if self.interrupted_text is None:
            return

        self.run["status"] = status
        self.run["interrupt_time"] = interrupt_time
        self.run["elapsed_time"] = td_format(interrupt_time - self.run["start_time"])

        try:
            self.bot.send_message(
                chat_id=self.chat_id,
                text=self.get_interrupted_text(),
                disable_notification=False,
                parse_mode=telegram.ParseMode.MARKDOWN,
            )
        except Exception as e:
            log = logging.getLogger("telegram-observer")
            log.warning(
                "failed to send interrupted_event message " "via telegram.", exc_info=e
            )

    def failed_event(self, fail_time, fail_trace):
        import telegram

        if self.failed_text is None:
            return

        self.run["fail_trace"] = fail_trace
        self.run["error"] = fail_trace[-1].strip()
        self.run["fail_time"] = fail_time
        self.run["elapsed_time"] = td_format(fail_time - self.run["start_time"])

        try:
            self.bot.send_message(
                chat_id=self.chat_id,
                text=self.get_failed_text(),
                disable_notification=False,
                parse_mode=telegram.ParseMode.MARKDOWN,
            )
        except Exception as e:
            log = logging.getLogger("telegram-observer")
            log.warning("failed to send failed_event message via telegram.", exc_info=e)



#!/usr/bin/env python
# coding=utf-8

from sacred.utils import iter_prefixes, join_paths


class ConfigSummary(dict):
    def __init__(
        self, added=(), modified=(), typechanged=(), ignored_fallbacks=(), docs=()
    ):
        super().__init__()
        self.added = set(added)
        self.modified = set(modified)  # TODO: test for this member
        self.typechanged = dict(typechanged)
        self.ignored_fallbacks = set(ignored_fallbacks)  # TODO: test
        self.docs = dict(docs)
        self.ensure_coherence()

    def update_from(self, config_mod, path=""):
        added = config_mod.added
        updated = config_mod.modified
        typechanged = config_mod.typechanged
        self.added &= {join_paths(path, a) for a in added}
        self.modified |= {join_paths(path, u) for u in updated}
        self.typechanged.update(
            {join_paths(path, k): v for k, v in typechanged.items()}
        )
        self.ensure_coherence()
        for k, v in config_mod.docs.items():
            if not self.docs.get(k, ""):
                self.docs[k] = v

    def update_add(self, config_mod, path=""):
        added = config_mod.added
        updated = config_mod.modified
        typechanged = config_mod.typechanged
        self.added |= {join_paths(path, a) for a in added}
        self.modified |= {join_paths(path, u) for u in updated}
        self.typechanged.update(
            {join_paths(path, k): v for k, v in typechanged.items()}
        )
        self.docs.update(
            {
                join_paths(path, k): v
                for k, v in config_mod.docs.items()
                if path == "" or k != "seed"
            }
        )
        self.ensure_coherence()

    def ensure_coherence(self):
        # make sure parent paths show up as updated appropriately
        self.modified |= {p for a in self.added for p in iter_prefixes(a)}
        self.modified |= {p for u in self.modified for p in iter_prefixes(u)}
        self.modified |= {p for t in self.typechanged for p in iter_prefixes(t)}

        # make sure there is no overlap
        self.added -= set(self.typechanged.keys())
        self.modified -= set(self.typechanged.keys())
        self.modified -= self.added



import json
import os
import os.path

from sacred.commandline_options import cli_option
from sacred.dependencies import get_digest
from sacred.observers.base import RunObserver
from sacred.serializer import flatten
import re
import socket

DEFAULT_S3_PRIORITY = 20


def _is_valid_bucket(bucket_name):
    # See https://docs.aws.amazon.com/awscloudtrail/latest/userguide/
    # cloudtrail-s3-bucket-naming-requirements.html
    if len(bucket_name) < 3 or len(bucket_name) > 63:
        return False

    labels = bucket_name.split(".")
    # A bucket name consists of "labels" separated by periods
    for label in labels:
        if len(label) == 0 or label[0] == "-" or label[-1] == "-":
            # Labels must be of nonzero length,
            # and cannot begin or end with a hyphen
            return False
        for char in label:
            # Labels can only contain digits, lowercase letters, or hyphens.
            # Anything else will fail here
            if not (char.isdigit() or char.islower() or char == "-"):
                return False
    try:
        # If a name is a valid IP address, it cannot be a bucket name
        socket.inet_aton(bucket_name)
    except socket.error:
        return True


def s3_join(*args):
    return "/".join(args)


class S3Observer(RunObserver):
    VERSION = "S3Observer-0.1.0"

    def __init__(
        self,
        bucket,
        basedir,
        resource_dir=None,
        source_dir=None,
        priority=DEFAULT_S3_PRIORITY,
        region=None,
    ):
        """Constructor for a S3Observer object.

        Run when the object is first created,
        before it's used within an experiment.

        Parameters
        ----------
        bucket
            The name of the bucket you want to store results in.
            Doesn't need to contain `s3://`, but needs to be a valid bucket name
        basedir
            The relative path inside your bucket where you want this experiment to store results
        resource_dir
            Where to store resources for this experiment. By
            default, will be <basedir>/_resources
        source_dir
            Where to store code sources for this experiment. By
            default, will be <basedir>/sources
        priority
            The priority to assign to this observer if
            multiple observers are present
        region
            The AWS region in which you want to create and access
            buckets. Needs to be either set here or configured in your AWS
        """
        import boto3

        if not _is_valid_bucket(bucket):
            raise ValueError(
                "Your chosen bucket name doesn't follow AWS bucket naming rules"
            )
        resource_dir = resource_dir or "/".join([basedir, "_resources"])
        source_dir = source_dir or "/".join([basedir, "_sources"])

        self.basedir = basedir
        self.bucket = bucket
        # Keeping the convention of referring to locations in S3 as `dir`
        # because that is a useful mental model and there isn't a better word
        self.resource_dir = resource_dir
        self.source_dir = source_dir
        self.priority = priority
        self.dir = None
        self.run_entry = None
        self.config = None
        self.info = None
        self.cout = ""
        self.cout_write_cursor = 0
        self.saved_metrics = {}
        if region is not None:
            self.region = region
            self.s3 = boto3.resource("s3", region_name=region)
        else:
            session = boto3.session.Session()
            if session.region_name is not None:
                self.region = session.region_name
                self.s3 = boto3.resource("s3")
            else:
                raise ValueError(
                    "You must either pass in an AWS region name, or have a "
                    "region name specified in your AWS config file"
                )

    def _objects_exist_in_dir(self, prefix):
        # This should be run after you've confirmed the bucket
        # exists, and will error out if it does not exist

        bucket = self.s3.Bucket(self.bucket)
        all_keys = [el.key for el in bucket.objects.filter(Prefix=prefix)]
        return len(all_keys) > 0

    def _bucket_exists(self):
        from botocore.errorfactory import ClientError

        try:
            self.s3.meta.client.head_bucket(Bucket=self.bucket)
        except ClientError as er:
            if er.response["Error"]["Code"] == "404":
                return False
        return True

    def _list_s3_subdirs(self, prefix=None):
        if prefix is None:
            prefix = self.basedir
        bucket = self.s3.Bucket(self.bucket)
        all_keys = [obj.key for obj in bucket.objects.filter(Prefix=prefix)]
        subdir_match = r"{prefix}\/(.*)\/".format(prefix=prefix)
        subdirs = []
        for key in all_keys:
            match_obj = re.match(subdir_match, key)
            if match_obj is None:
                continue
            else:
                subdirs.append(match_obj.groups()[0])
        distinct_subdirs = set(subdirs)
        return list(distinct_subdirs)

    def _create_bucket(self):
        bucket_response = self.s3.create_bucket(
            Bucket=self.bucket,
            CreateBucketConfiguration={"LocationConstraint": self.region},
        )
        return bucket_response

    def _determine_run_dir(self, _id):
        if _id is None:
            bucket_exists = self._bucket_exists()
            if not bucket_exists:
                self._create_bucket()
                bucket_path_subdirs = []
            else:
                bucket_path_subdirs = self._list_s3_subdirs()

            if not bucket_path_subdirs:
                max_run_id = 0
            else:
                integer_directories = [
                    int(d) for d in bucket_path_subdirs if d.isdigit()
                ]
                if not integer_directories:
                    max_run_id = 0
                else:
                    # If there are directories under basedir that aren't
                    # numeric run directories, ignore those
                    max_run_id = max(integer_directories)

            _id = max_run_id + 1

        self.dir = s3_join(self.basedir, str(_id))
        if self._objects_exist_in_dir(self.dir):
            raise FileExistsError("S3 dir at {} already exists".format(self.dir))
        return _id

    def queued_event(
        self, ex_info, command, host_info, queue_time, config, meta_info, _id
    ):
        _id = self._determine_run_dir(_id)

        self.run_entry = {
            "experiment": dict(ex_info),
            "command": command,
            "host": dict(host_info),
            "meta": meta_info,
            "status": "QUEUED",
        }
        self.config = config
        self.info = {}

        self.save_json(self.run_entry, "run.json")
        self.save_json(self.config, "config.json")

        for s, m in ex_info["sources"]:
            self.save_file(s)

        return _id

    def save_sources(self, ex_info):
        base_dir = ex_info["base_dir"]
        source_info = []
        for s, m in ex_info["sources"]:
            abspath = os.path.join(base_dir, s)
            store_path, md5sum = self.find_or_save(abspath, self.source_dir)
            source_info.append([s, os.path.relpath(store_path, self.basedir)])
        return source_info

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):

        _id = self._determine_run_dir(_id)

        ex_info["sources"] = self.save_sources(ex_info)

        self.run_entry = {
            "experiment": dict(ex_info),
            "command": command,
            "host": dict(host_info),
            "start_time": start_time.isoformat(),
            "meta": meta_info,
            "status": "RUNNING",
            "resources": [],
            "artifacts": [],
            "heartbeat": None,
        }
        self.config = config
        self.info = {}
        self.cout = ""
        self.cout_write_cursor = 0

        self.save_json(self.run_entry, "run.json")
        self.save_json(self.config, "config.json")
        self.save_cout()

        return _id

    def find_or_save(self, filename, store_dir):
        source_name, ext = os.path.splitext(os.path.basename(filename))
        md5sum = get_digest(filename)
        store_name = source_name + "_" + md5sum + ext
        store_path = s3_join(store_dir, store_name)
        if len(self._list_s3_subdirs(prefix=store_path)) == 0:
            self.save_file(filename, store_path)
        return store_path, md5sum

    def put_data(self, key, binary_data):
        self.s3.Object(self.bucket, key).put(Body=binary_data)

    def save_json(self, obj, filename):
        key = s3_join(self.dir, filename)
        self.put_data(key, json.dumps(flatten(obj), sort_keys=True, indent=2))

    def save_file(self, filename, target_name=None):
        target_name = target_name or os.path.basename(filename)
        key = s3_join(self.dir, target_name)
        self.put_data(key, open(filename, "rb"))

    def save_directory(self, source_dir, target_name):
        import boto3

        # Stolen from:
        # https://github.com/boto/boto3/issues/358#issuecomment-346093506
        target_name = target_name or os.path.basename(source_dir)
        all_files = []
        for root, dirs, files in os.walk(source_dir):
            all_files += [os.path.join(root, f) for f in files]
        s3_resource = boto3.resource("s3")

        for filename in all_files:
            file_location = s3_join(
                self.dir, target_name, os.path.relpath(filename, source_dir)
            )
            s3_resource.Object(self.bucket, file_location).put(
                Body=open(filename, "rb")
            )

    def save_cout(self):
        binary_data = self.cout[self.cout_write_cursor :].encode("utf-8")
        key = s3_join(self.dir, "cout.txt")
        self.put_data(key, binary_data)
        self.cout_write_cursor = len(self.cout)

    def heartbeat_event(self, info, captured_out, beat_time, result):
        self.info = info
        self.run_entry["heartbeat"] = beat_time.isoformat()
        self.run_entry["result"] = result
        self.cout = captured_out
        self.save_cout()
        self.save_json(self.run_entry, "run.json")
        if self.info:
            self.save_json(self.info, "info.json")

    def completed_event(self, stop_time, result):
        self.run_entry["stop_time"] = stop_time.isoformat()
        self.run_entry["result"] = result
        self.run_entry["status"] = "COMPLETED"

        self.save_json(self.run_entry, "run.json")

    def interrupted_event(self, interrupt_time, status):
        self.run_entry["stop_time"] = interrupt_time.isoformat()
        self.run_entry["status"] = status
        self.save_json(self.run_entry, "run.json")

    def failed_event(self, fail_time, fail_trace):
        self.run_entry["stop_time"] = fail_time.isoformat()
        self.run_entry["status"] = "FAILED"
        self.run_entry["fail_trace"] = fail_trace
        self.save_json(self.run_entry, "run.json")

    def resource_event(self, filename):
        store_path, md5sum = self.find_or_save(filename, self.resource_dir)
        self.run_entry["resources"].append([filename, store_path])
        self.save_json(self.run_entry, "run.json")

    def artifact_event(self, name, filename, metadata=None, content_type=None):
        self.save_file(filename, name)
        self.run_entry["artifacts"].append(name)
        self.save_json(self.run_entry, "run.json")

    def log_metrics(self, metrics_by_name, info):
        """Store new measurements into metrics.json."""
        for metric_name, metric_ptr in metrics_by_name.items():

            if metric_name not in self.saved_metrics:
                self.saved_metrics[metric_name] = {
                    "values": [],
                    "steps": [],
                    "timestamps": [],
                }

            self.saved_metrics[metric_name]["values"] += metric_ptr["values"]
            self.saved_metrics[metric_name]["steps"] += metric_ptr["steps"]

            timestamps_norm = [ts.isoformat() for ts in metric_ptr["timestamps"]]
            self.saved_metrics[metric_name]["timestamps"] += timestamps_norm

        self.save_json(self.saved_metrics, "metrics.json")

    def __eq__(self, other):
        if isinstance(other, S3Observer):
            return self.bucket == other.bucket and self.basedir == other.basedir
        else:
            return False


@cli_option("-S", "--s3")
def s3_option(args, run):
    """Add a S3 File observer to the experiment.

    The argument value should be `s3://<bucket>/path/to/exp`.
    """
    match_obj = re.match(r"s3:\/\/([^\/]*)\/(.*)", args)
    if match_obj is None or len(match_obj.groups()) != 2:
        raise ValueError(
            "Valid bucket specification not found. "
            "Enter bucket and directory path like: "
            "s3://<bucket>/path/to/exp"
        )
    bucket, basedir = match_obj.groups()
    run.observers.append(S3Observer(bucket=bucket, basedir=basedir))



#!/usr/bin/env python
# coding=utf-8

from sacred.config.config_summary import ConfigSummary
from sacred.config.utils import (
    dogmatize,
    normalize_or_die,
    undogmatize,
    recursive_fill_in,
)


class ConfigDict:
    def __init__(self, d):
        self._conf = normalize_or_die(d)

    def __call__(self, fixed=None, preset=None, fallback=None):
        result = dogmatize(fixed or {})
        recursive_fill_in(result, self._conf)
        recursive_fill_in(result, preset or {})
        added = result.revelation()
        config_summary = ConfigSummary(added, result.modified, result.typechanges)
        config_summary.update(undogmatize(result))
        return config_summary



from sacred.observers.base import RunObserver
from sacred.observers.file_storage import FileStorageObserver
from sacred.observers.mongo import MongoObserver, QueuedMongoObserver
from sacred.observers.sql import SqlObserver
from sacred.observers.tinydb_hashfs import TinyDbObserver, TinyDbReader
from sacred.observers.slack import SlackObserver
from sacred.observers.telegram_obs import TelegramObserver
from sacred.observers.s3_observer import S3Observer
from sacred.observers.queue import QueueObserver


__all__ = (
    "FileStorageObserver",
    "RunObserver",
    "MongoObserver",
    "QueuedMongoObserver",
    "SqlObserver",
    "TinyDbObserver",
    "TinyDbReader",
    "SlackObserver",
    "TelegramObserver",
    "S3Observer",
    "QueueObserver",
)



"""The Experiment class, which is central to sacred."""
import inspect
import os.path
import sys
import warnings
from collections import OrderedDict
from typing import Sequence, Optional, List

from docopt import docopt, printable_usage

from sacred import SETTINGS
from sacred.arg_parser import format_usage, get_config_updates
from sacred import commandline_options
from sacred.commandline_options import CLIOption
from sacred.commands import (
    help_for_command,
    print_config,
    print_dependencies,
    save_config,
    print_named_configs,
)
from sacred.observers.file_storage import file_storage_option
from sacred.observers.s3_observer import s3_option
from sacred.config.signature import Signature
from sacred.ingredient import Ingredient
from sacred.initialize import create_run
from sacred.observers.sql import sql_option
from sacred.observers.tinydb_hashfs import tiny_db_option
from sacred.run import Run
from sacred.host_info import check_additional_host_info, HostInfoGetter
from sacred.utils import (
    print_filtered_stacktrace,
    ensure_wellformed_argv,
    SacredError,
    format_sacred_error,
    PathType,
    get_inheritors,
)
from sacred.observers.mongo import mongo_db_option

__all__ = ("Experiment",)


class Experiment(Ingredient):
    """
    The central class for each experiment in Sacred.

    It manages the configuration, the main function, captured methods,
    observers, commands, and further ingredients.

    An Experiment instance should be created as one of the first
    things in any experiment-file.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        ingredients: Sequence[Ingredient] = (),
        interactive: bool = False,
        base_dir: Optional[PathType] = None,
        additional_host_info: Optional[List[HostInfoGetter]] = None,
        additional_cli_options: Optional[Sequence[CLIOption]] = None,
        save_git_info: bool = True,
    ):
        """
        Create a new experiment with the given name and optional ingredients.

        Parameters
        ----------
        name
            Optional name of this experiment, defaults to the filename.
            (Required in interactive mode)

        ingredients : list[sacred.Ingredient], optional
            A list of ingredients to be used with this experiment.

        interactive
            If set to True will allow the experiment to be run in interactive
            mode (e.g. IPython or Jupyter notebooks).
            However, this mode is discouraged since it won't allow storing the
            source-code or reliable reproduction of the runs.

        base_dir
            Optional full path to the base directory of this experiment. This
            will set the scope for automatic source file discovery.

        additional_host_info
            Optional dictionary containing as keys the names of the pieces of
            host info you want to collect, and as
            values the functions collecting those pieces of information.

        save_git_info:
            Optionally save the git commit hash and the git state
            (clean or dirty) for all source files. This requires the GitPython
            package.
        """
        self.additional_host_info = additional_host_info or []
        check_additional_host_info(self.additional_host_info)
        self.additional_cli_options = additional_cli_options or []
        self.all_cli_options = (
            gather_command_line_options() + self.additional_cli_options
        )
        caller_globals = inspect.stack()[1][0].f_globals
        if name is None:
            if interactive:
                raise RuntimeError("name is required in interactive mode.")
            mainfile = caller_globals.get("__file__")
            if mainfile is None:
                raise RuntimeError(
                    "No main-file found. Are you running in "
                    "interactive mode? If so please provide a "
                    "name and set interactive=True."
                )
            name = os.path.basename(mainfile)
            if name.endswith(".py"):
                name = name[:-3]
            elif name.endswith(".pyc"):
                name = name[:-4]
        super().__init__(
            path=name,
            ingredients=ingredients,
            interactive=interactive,
            base_dir=base_dir,
            _caller_globals=caller_globals,
            save_git_info=save_git_info,
        )
        self.default_command = None
        self.command(print_config, unobserved=True)
        self.command(print_dependencies, unobserved=True)
        self.command(save_config, unobserved=True)
        self.command(print_named_configs(self), unobserved=True)
        self.observers = []
        self.current_run = None
        self.captured_out_filter = None
        """Filter function to be applied to captured output of a run"""
        self.option_hooks = []

    # =========================== Decorators ==================================

    def main(self, function):
        """
        Decorator to define the main function of the experiment.

        The main function of an experiment is the default command that is being
        run when no command is specified, or when calling the run() method.

        Usually it is more convenient to use ``automain`` instead.
        """
        captured = self.command(function)
        self.default_command = captured.__name__
        return captured

    def automain(self, function):
        """
        Decorator that defines *and runs* the main function of the experiment.

        The decorated function is marked as the default command for this
        experiment, and the command-line interface is automatically run when
        the file is executed.

        The method decorated by this should be last in the file because is
        equivalent to:

        Example
        -------
        ::

            @ex.main
            def my_main():
                pass

            if __name__ == '__main__':
                ex.run_commandline()
        """
        captured = self.main(function)
        if function.__module__ == "__main__":
            # Ensure that automain is not used in interactive mode.
            import inspect

            main_filename = inspect.getfile(function)
            if main_filename == "<stdin>" or (
                main_filename.startswith("<ipython-input-")
                and main_filename.endswith(">")
            ):
                raise RuntimeError(
                    "Cannot use @ex.automain decorator in "
                    "interactive mode. Use @ex.main instead."
                )

            self.run_commandline()
        return captured

    def option_hook(self, function):
        """
        Decorator for adding an option hook function.

        An option hook is a function that is called right before a run
        is created. It receives (and potentially modifies) the options
        dictionary. That is, the dictionary of commandline options used for
        this run.

        Notes
        -----
        The decorated function MUST have an argument called options.

        The options also contain ``'COMMAND'`` and ``'UPDATE'`` entries,
        but changing them has no effect. Only modification on
        flags (entries starting with ``'--'``) are considered.
        """
        sig = Signature(function)
        if "options" not in sig.arguments:
            raise KeyError(
                "option_hook functions must have an argument called"
                " 'options', but got {}".format(sig.arguments)
            )
        self.option_hooks.append(function)
        return function

    # =========================== Public Interface ============================

    def get_usage(self, program_name=None):
        """Get the commandline usage string for this experiment."""
        program_name = os.path.relpath(
            program_name or sys.argv[0] or "Dummy", self.base_dir
        )
        commands = OrderedDict(self.gather_commands())
        long_usage = format_usage(
            program_name, self.doc, commands, self.all_cli_options
        )
        # internal usage is a workaround because docopt cannot handle spaces
        # in program names. So for parsing we use 'dummy' as the program name.
        # for printing help etc. we want to use the actual program name.
        internal_usage = format_usage("dummy", self.doc, commands, self.all_cli_options)
        short_usage = printable_usage(long_usage)
        return short_usage, long_usage, internal_usage

    def run(
        self,
        command_name: Optional[str] = None,
        config_updates: Optional[dict] = None,
        named_configs: Sequence[str] = (),
        info: Optional[dict] = None,
        meta_info: Optional[dict] = None,
        options: Optional[dict] = None,
    ) -> Run:
        """
        Run the main function of the experiment or a given command.

        Parameters
        ----------
        command_name
            Name of the command to be run. Defaults to main function.

        config_updates
            Changes to the configuration as a nested dictionary

        named_configs
            list of names of named_configs to use

        info
            Additional information for this run.

        meta_info
            Additional meta information for this run.

        options
            Dictionary of options to use

        Returns
        -------
        The Run object corresponding to the finished run.
        """
        run = self._create_run(
            command_name, config_updates, named_configs, info, meta_info, options
        )
        run()
        return run

    def run_commandline(self, argv=None) -> Optional[Run]:
        """
        Run the command-line interface of this experiment.

        If ``argv`` is omitted it defaults to ``sys.argv``.

        Parameters
        ----------
        argv
            Command-line as string or list of strings like ``sys.argv``.

        Returns
        -------
        The Run object corresponding to the finished run.

        """
        argv = ensure_wellformed_argv(argv)
        short_usage, usage, internal_usage = self.get_usage()
        args = docopt(internal_usage, [str(a) for a in argv[1:]], help=False)

        cmd_name = args.get("COMMAND") or self.default_command
        config_updates, named_configs = get_config_updates(args["UPDATE"])

        err = self._check_command(cmd_name)
        if not args["help"] and err:
            print(short_usage)
            print(err)
            sys.exit(1)

        if self._handle_help(args, usage):
            sys.exit()

        try:
            return self.run(
                cmd_name,
                config_updates,
                named_configs,
                info={},
                meta_info={},
                options=args,
            )
        except Exception as e:
            if self.current_run:
                debug = self.current_run.debug
            else:
                # The usual command line options are applied after the run
                # object is built completely. Some exceptions (e.g.
                # ConfigAddedError) are raised before this. In these cases,
                # the debug flag must be checked manually.
                debug = args.get("--debug", False)

            if debug:
                # Debug: Don't change behavior, just re-raise exception
                raise
            elif self.current_run and self.current_run.pdb:
                # Print exception and attach pdb debugger
                import traceback
                import pdb

                traceback.print_exception(*sys.exc_info())
                pdb.post_mortem()
            else:
                # Handle pretty printing of exceptions. This includes
                # filtering the stacktrace and printing the usage, as
                # specified by the exceptions attributes
                if isinstance(e, SacredError):
                    print(format_sacred_error(e, short_usage), file=sys.stderr)
                else:
                    print_filtered_stacktrace()
                sys.exit(1)

    def open_resource(self, filename: PathType, mode: str = "r"):
        """Open a file and also save it as a resource.

        Opens a file, reports it to the observers as a resource, and returns
        the opened file.

        In Sacred terminology a resource is a file that the experiment needed
        to access during a run. In case of a MongoObserver that means making
        sure the file is stored in the database (but avoiding duplicates) along
        its path and md5 sum.

        This function can only be called during a run, and just calls the
        :py:meth:`sacred.run.Run.open_resource` method.

        Parameters
        ----------
        filename
            name of the file that should be opened
        mode
            mode that file will be open

        Returns
        -------
        The opened file-object.
        """
        assert self.current_run is not None, "Can only be called during a run."
        return self.current_run.open_resource(filename, mode)

    def add_resource(self, filename: PathType) -> None:
        """Add a file as a resource.

        In Sacred terminology a resource is a file that the experiment needed
        to access during a run. In case of a MongoObserver that means making
        sure the file is stored in the database (but avoiding duplicates) along
        its path and md5 sum.

        This function can only be called during a run, and just calls the
        :py:meth:`sacred.run.Run.add_resource` method.

        Parameters
        ----------
        filename
            name of the file to be stored as a resource
        """
        assert self.current_run is not None, "Can only be called during a run."
        self.current_run.add_resource(filename)

    def add_artifact(
        self,
        filename: PathType,
        name: Optional[str] = None,
        metadata: Optional[dict] = None,
        content_type: Optional[str] = None,
    ) -> None:
        """Add a file as an artifact.

        In Sacred terminology an artifact is a file produced by the experiment
        run. In case of a MongoObserver that means storing the file in the
        database.

        This function can only be called during a run, and just calls the
        :py:meth:`sacred.run.Run.add_artifact` method.

        Parameters
        ----------
        filename
            name of the file to be stored as artifact
        name
            optionally set the name of the artifact.
            Defaults to the relative file-path.
        metadata
            optionally attach metadata to the artifact.
            This only has an effect when using the MongoObserver.
        content_type
            optionally attach a content-type to the artifact.
            This only has an effect when using the MongoObserver.
        """
        assert self.current_run is not None, "Can only be called during a run."
        self.current_run.add_artifact(filename, name, metadata, content_type)

    @property
    def info(self) -> dict:
        """Access the info-dict for storing custom information.

        Only works during a run and is essentially a shortcut to:

        Example
        -------
        ::

            @ex.capture
            def my_captured_function(_run):
                # [...]
                _run.info   # == ex.info
        """
        return self.current_run.info

    def log_scalar(self, name: str, value: float, step: Optional[int] = None) -> None:
        """
        Add a new measurement.

        The measurement will be processed by the MongoDB* observer
        during a heartbeat event.
        Other observers are not yet supported.


        Parameters
        ----------
        name
            The name of the metric, e.g. training.loss
        value
            The measured value
        step
            The step number (integer), e.g. the iteration number
            If not specified, an internal counter for each metric
            is used, incremented by one.
        """
        # Method added in change https://github.com/chovanecm/sacred/issues/4
        # The same as Run.log_scalar
        self.current_run.log_scalar(name, value, step)

    def post_process_name(self, name, ingredient):
        if ingredient == self:
            # Removes the experiment's path (prefix) from the names
            # of the gathered items. This means that, for example,
            # 'experiment.print_config' becomes 'print_config'.
            return name[len(self.path) + 1 :]
        return name

    def get_default_options(self) -> dict:
        """Get a dictionary of default options as used with run.

        Returns
        -------
        A dictionary containing option keys of the form '--beat_interval'.
        Their values are boolean if the option is a flag, otherwise None or
        its default value.

        """
        default_options = {}
        for option in self.all_cli_options:
            if isinstance(option, CLIOption):
                if option.is_flag:
                    default_value = False
                else:
                    default_value = None
            else:  # legacy, should be removed later on.
                if option.arg is None:
                    default_value = False
                else:
                    default_value = None
            default_options[option.get_flag()] = default_value

        return default_options

    # =========================== Internal Interface ==========================

    def _create_run(
        self,
        command_name=None,
        config_updates=None,
        named_configs=(),
        info=None,
        meta_info=None,
        options=None,
    ):
        command_name = command_name or self.default_command
        if command_name is None:
            raise RuntimeError(
                "No command found to be run. Specify a command "
                "or define a main function."
            )

        default_options = self.get_default_options()
        if options:
            default_options.update(options)
        options = default_options

        # call option hooks
        for oh in self.option_hooks:
            oh(options=options)

        run = create_run(
            self,
            command_name,
            config_updates,
            named_configs=named_configs,
            force=options.get(commandline_options.force_option.get_flag(), False),
            log_level=options.get(commandline_options.loglevel_option.get_flag(), None),
        )
        if info is not None:
            run.info.update(info)

        run.meta_info["command"] = command_name
        run.meta_info["options"] = options

        if meta_info:
            run.meta_info.update(meta_info)

        options_list = gather_command_line_options() + self.additional_cli_options
        for option in options_list:
            option_value = options.get(option.get_flag(), False)
            if option_value:
                option.apply(option_value, run)

        self.current_run = run
        return run

    def _check_command(self, cmd_name):
        commands = dict(self.gather_commands())
        if cmd_name is not None and cmd_name not in commands:
            return 'Error: Command "{}" not found. Available commands are: ' "{}".format(
                cmd_name, ", ".join(commands.keys())
            )

        if cmd_name is None:
            return (
                "Error: No command found to be run. Specify a command"
                " or define main function. Available commands"
                " are: {}".format(", ".join(commands.keys()))
            )

    def _handle_help(self, args, usage):
        if args["help"] or args["--help"]:
            if args["COMMAND"] is None:
                print(usage)
                return True
            else:
                commands = dict(self.gather_commands())
                print(help_for_command(commands[args["COMMAND"]]))
                return True
        return False


def gather_command_line_options(filter_disabled=None):
    """Get a sorted list of all CommandLineOption subclasses."""
    if filter_disabled is None:
        filter_disabled = not SETTINGS.COMMAND_LINE.SHOW_DISABLED_OPTIONS

    options = []
    for opt in get_inheritors(commandline_options.CommandLineOption):
        warnings.warn(
            "Subclassing `CommandLineOption` is deprecated. Please "
            "use the `sacred.cli_option` decorator and pass the function "
            "to the Experiment constructor."
        )
        if filter_disabled and not opt._enabled:
            continue
        options.append(opt)

    options += DEFAULT_COMMAND_LINE_OPTIONS

    return sorted(options, key=commandline_options.get_name)


DEFAULT_COMMAND_LINE_OPTIONS = [
    s3_option,
    commandline_options.pdb_option,
    commandline_options.debug_option,
    file_storage_option,
    commandline_options.loglevel_option,
    mongo_db_option,
    sql_option,
    commandline_options.capture_option,
    commandline_options.help_option,
    commandline_options.print_config_option,
    commandline_options.name_option,
    commandline_options.priority_option,
    commandline_options.unobserved_option,
    commandline_options.beat_interval_option,
    commandline_options.queue_option,
    commandline_options.force_option,
    commandline_options.comment_option,
    commandline_options.enforce_clean_option,
    tiny_db_option,
]



#!/usr/bin/env python
# coding=utf-8

from sacred.config.config_dict import ConfigDict
from sacred.config.config_scope import ConfigScope
from sacred.config.config_files import load_config_file, save_config_file
from sacred.config.captured_function import create_captured_function
from sacred.config.utils import chain_evaluate_config_scopes, dogmatize, undogmatize

__all__ = (
    "ConfigDict",
    "ConfigScope",
    "load_config_file",
    "save_config_file",
    "create_captured_function",
    "chain_evaluate_config_scopes",
    "dogmatize",
    "undogmatize",
)



#!/usr/bin/env python
# coding=utf-8

import jsonpickle as json
from jsonpickle.handlers import BaseHandler
import json as _json

from sacred import optional as opt

__all__ = ("flatten", "restore")


# class DatetimeHandler(BaseHandler):
#     def restore(self, obj):
#        return datetime.datetime.strptime(obj['date'], "%Y-%m-%dT%H:%M:%S.%f")
#
#     def flatten(self, obj, data):
#         data['date'] = obj.isoformat()
#         return data
#
# DatetimeHandler.handles(datetime.datetime)

if opt.has_numpy:
    np = opt.np

    class NumpyArrayHandler(BaseHandler):
        def flatten(self, obj, data):
            data["values"] = obj.tolist()
            data["dtype"] = str(obj.dtype)
            return data

        def restore(self, obj):
            return opt.np.array(obj["values"], dtype=obj["dtype"])

    class NumpyGenericHandler(BaseHandler):
        def flatten(self, obj, data):
            return obj.item()

        def restore(self, obj):
            return obj

    NumpyArrayHandler.handles(np.ndarray)
    for t in [
        np.bool_,
        np.int_,
        np.float_,
        np.intc,
        np.intp,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
        np.float16,
        np.float32,
        np.float64,
    ]:
        NumpyGenericHandler.handles(t)


if opt.has_pandas:
    import pandas as pd

    class PandasDataframeHandler(BaseHandler):
        def flatten(self, obj, data):
            # TODO: this is slow
            data["values"] = json.loads(obj.to_json())
            data["dtypes"] = {k: str(v) for k, v in dict(obj.dtypes).items()}
            return data

        def restore(self, obj):
            # TODO: get rid of unnecessary json.dumps
            return pd.read_json(json.dumps(obj["values"]), dtype=obj["dtypes"])

    PandasDataframeHandler.handles(pd.DataFrame)

json.set_encoder_options("simplejson", sort_keys=True, indent=4)
json.set_encoder_options("demjson", compactly=False)


def flatten(obj):
    return _json.loads(json.encode(obj, keys=True))


def restore(flat):
    return json.decode(_json.dumps(flat), keys=True)



#!/usr/bin/env python
# coding=utf-8

__all__ = ("RunObserver", "td_format")


class RunObserver:
    """Defines the interface for all run observers."""

    priority = 0

    def queued_event(
        self, ex_info, command, host_info, queue_time, config, meta_info, _id
    ):
        pass

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):
        pass

    def heartbeat_event(self, info, captured_out, beat_time, result):
        pass

    def completed_event(self, stop_time, result):
        pass

    def interrupted_event(self, interrupt_time, status):
        pass

    def failed_event(self, fail_time, fail_trace):
        pass

    def resource_event(self, filename):
        pass

    def artifact_event(self, name, filename, metadata=None, content_type=None):
        pass

    def log_metrics(self, metrics_by_name, info):
        pass

    def join(self):
        pass


# http://stackoverflow.com/questions/538666/python-format-timedelta-to-string
def td_format(td_object):
    seconds = int(td_object.total_seconds())
    if seconds == 0:
        return "less than a second"

    periods = [
        ("year", 60 * 60 * 24 * 365),
        ("month", 60 * 60 * 24 * 30),
        ("day", 60 * 60 * 24),
        ("hour", 60 * 60),
        ("minute", 60),
        ("second", 1),
    ]

    strings = []
    for period_name, period_seconds in periods:
        if seconds >= period_seconds:
            period_value, seconds = divmod(seconds, period_seconds)
            if period_value == 1:
                strings.append("%s %s" % (period_value, period_name))
            else:
                strings.append("%s %ss" % (period_value, period_name))

    return ", ".join(strings)



#!/usr/bin/env python
# coding=utf-8
"""
The main module of sacred.

It provides access to the two main classes Experiment and Ingredient.
"""

from sacred.__about__ import __version__, __author__, __author_email__, __url__
from sacred.settings import SETTINGS
from sacred.experiment import Experiment
from sacred.ingredient import Ingredient
from sacred import observers
from sacred.host_info import host_info_getter, host_info_gatherer
from sacred.commandline_options import cli_option


__all__ = (
    "Experiment",
    "Ingredient",
    "observers",
    "host_info_getter",
    "__version__",
    "__author__",
    "__author_email__",
    "__url__",
    "SETTINGS",
    "host_info_gatherer",
    "cli_option",
)



#!/usr/bin/env python
# coding=utf-8

from pathlib import Path

from setuptools import setup

classifiers = """
Development Status :: 5 - Production/Stable
Intended Audience :: Science/Research
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Utilities
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Artificial Intelligence
Topic :: Software Development :: Libraries :: Python Modules
License :: OSI Approved :: MIT License
"""

try:
    from sacred import __about__

    about = __about__.__dict__
except ImportError:
    # installing - dependencies are not there yet
    # Manually extract the __about__
    about = dict()
    exec(open("sacred/__about__.py").read(), about)


setup(
    name="sacred",
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    packages=["sacred", "sacred.observers", "sacred.config", "sacred.stflow"],
    scripts=[],
    install_requires=Path("requirements.txt").read_text().splitlines(),
    tests_require=["mock>=0.8, <3.0", "pytest==4.3.0"],
    classifiers=list(filter(None, classifiers.split("\n"))),
    description="Facilitates automated and reproducible experimental research",
    long_description=Path("README.rst").read_text(),
)



import hashlib
import json
import os

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base

from sacred.dependencies import get_digest
from sacred.serializer import restore


Base = declarative_base()


class Source(Base):
    __tablename__ = "source"

    @classmethod
    def get_or_create(cls, filename, md5sum, session, basedir):
        instance = (
            session.query(cls).filter_by(filename=filename, md5sum=md5sum).first()
        )
        if instance:
            return instance
        full_path = os.path.join(basedir, filename)
        md5sum_ = get_digest(full_path)
        assert md5sum_ == md5sum, "found md5 mismatch for {}: {} != {}".format(
            filename, md5sum, md5sum_
        )
        with open(full_path, "r") as f:
            return cls(filename=filename, md5sum=md5sum, content=f.read())

    source_id = sa.Column(sa.Integer, primary_key=True)
    filename = sa.Column(sa.String(256))
    md5sum = sa.Column(sa.String(32))
    content = sa.Column(sa.Text)

    def to_json(self):
        return {"filename": self.filename, "md5sum": self.md5sum}


class Dependency(Base):
    __tablename__ = "dependency"

    @classmethod
    def get_or_create(cls, dep, session):
        name, _, version = dep.partition("==")
        instance = session.query(cls).filter_by(name=name, version=version).first()
        if instance:
            return instance
        return cls(name=name, version=version)

    dependency_id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(32))
    version = sa.Column(sa.String(16))

    def to_json(self):
        return "{}=={}".format(self.name, self.version)


class Artifact(Base):
    __tablename__ = "artifact"

    @classmethod
    def create(cls, name, filename):
        with open(filename, "rb") as f:
            return cls(filename=name, content=f.read())

    artifact_id = sa.Column(sa.Integer, primary_key=True)
    filename = sa.Column(sa.String(64))
    content = sa.Column(sa.LargeBinary)

    run_id = sa.Column(sa.String(24), sa.ForeignKey("run.run_id"))
    run = sa.orm.relationship("Run", backref=sa.orm.backref("artifacts"))

    def to_json(self):
        return {"_id": self.artifact_id, "filename": self.filename}


class Resource(Base):
    __tablename__ = "resource"

    @classmethod
    def get_or_create(cls, filename, session):
        md5sum = get_digest(filename)
        instance = (
            session.query(cls).filter_by(filename=filename, md5sum=md5sum).first()
        )
        if instance:
            return instance
        with open(filename, "rb") as f:
            return cls(filename=filename, md5sum=md5sum, content=f.read())

    resource_id = sa.Column(sa.Integer, primary_key=True)
    filename = sa.Column(sa.String(256))
    md5sum = sa.Column(sa.String(32))
    content = sa.Column(sa.LargeBinary)

    def to_json(self):
        return {"filename": self.filename, "md5sum": self.md5sum}


class Host(Base):
    __tablename__ = "host"

    @classmethod
    def get_or_create(cls, host_info, session):
        h = dict(
            hostname=host_info["hostname"],
            cpu=host_info["cpu"],
            os=host_info["os"][0],
            os_info=host_info["os"][1],
            python_version=host_info["python_version"],
        )

        return session.query(cls).filter_by(**h).first() or cls(**h)

    host_id = sa.Column(sa.Integer, primary_key=True)
    cpu = sa.Column(sa.String(64))
    hostname = sa.Column(sa.String(64))
    os = sa.Column(sa.String(16))
    os_info = sa.Column(sa.String(64))
    python_version = sa.Column(sa.String(16))

    def to_json(self):
        return {
            "cpu": self.cpu,
            "hostname": self.hostname,
            "os": [self.os, self.os_info],
            "python_version": self.python_version,
        }


experiment_source_association = sa.Table(
    "experiments_sources",
    Base.metadata,
    sa.Column("experiment_id", sa.Integer, sa.ForeignKey("experiment.experiment_id")),
    sa.Column("source_id", sa.Integer, sa.ForeignKey("source.source_id")),
)

experiment_dependency_association = sa.Table(
    "experiments_dependencies",
    Base.metadata,
    sa.Column("experiment_id", sa.Integer, sa.ForeignKey("experiment.experiment_id")),
    sa.Column("dependency_id", sa.Integer, sa.ForeignKey("dependency.dependency_id")),
)


class Experiment(Base):
    __tablename__ = "experiment"

    @classmethod
    def get_or_create(cls, ex_info, session):
        name = ex_info["name"]
        # Compute a MD5sum of the ex_info to determine its uniqueness
        h = hashlib.md5()
        h.update(json.dumps(ex_info).encode())
        md5 = h.hexdigest()
        instance = session.query(cls).filter_by(name=name, md5sum=md5).first()
        if instance:
            return instance

        dependencies = [
            Dependency.get_or_create(d, session) for d in ex_info["dependencies"]
        ]
        sources = [
            Source.get_or_create(s, md5sum, session, ex_info["base_dir"])
            for s, md5sum in ex_info["sources"]
        ]

        return cls(
            name=name,
            dependencies=dependencies,
            sources=sources,
            md5sum=md5,
            base_dir=ex_info["base_dir"],
        )

    experiment_id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(32))
    md5sum = sa.Column(sa.String(32))
    base_dir = sa.Column(sa.String(64))
    sources = sa.orm.relationship(
        "Source", secondary=experiment_source_association, backref="experiments"
    )
    dependencies = sa.orm.relationship(
        "Dependency", secondary=experiment_dependency_association, backref="experiments"
    )

    def to_json(self):
        return {
            "name": self.name,
            "base_dir": self.base_dir,
            "sources": [s.to_json() for s in self.sources],
            "dependencies": [d.to_json() for d in self.dependencies],
        }


run_resource_association = sa.Table(
    "runs_resources",
    Base.metadata,
    sa.Column("run_id", sa.String(24), sa.ForeignKey("run.run_id")),
    sa.Column("resource_id", sa.Integer, sa.ForeignKey("resource.resource_id")),
)


class Run(Base):
    __tablename__ = "run"
    id = sa.Column(sa.Integer, primary_key=True)

    run_id = sa.Column(sa.String(24), unique=True)

    command = sa.Column(sa.String(64))

    # times
    start_time = sa.Column(sa.DateTime)
    heartbeat = sa.Column(sa.DateTime)
    stop_time = sa.Column(sa.DateTime)
    queue_time = sa.Column(sa.DateTime)

    # meta info
    priority = sa.Column(sa.Float)
    comment = sa.Column(sa.Text)

    fail_trace = sa.Column(sa.Text)

    # Captured out
    # TODO: move to separate table?
    captured_out = sa.Column(sa.Text)

    # Configuration & info
    # TODO: switch type to json if possible
    config = sa.Column(sa.Text)
    info = sa.Column(sa.Text)

    status = sa.Column(
        sa.Enum(
            "RUNNING",
            "COMPLETED",
            "INTERRUPTED",
            "TIMEOUT",
            "FAILED",
            name="status_enum",
        )
    )

    host_id = sa.Column(sa.Integer, sa.ForeignKey("host.host_id"))
    host = sa.orm.relationship("Host", backref=sa.orm.backref("runs"))

    experiment_id = sa.Column(sa.Integer, sa.ForeignKey("experiment.experiment_id"))
    experiment = sa.orm.relationship("Experiment", backref=sa.orm.backref("runs"))

    # artifacts = backref
    resources = sa.orm.relationship(
        "Resource", secondary=run_resource_association, backref="runs"
    )

    result = sa.Column(sa.Float)

    def to_json(self):
        return {
            "_id": self.run_id,
            "command": self.command,
            "start_time": self.start_time,
            "heartbeat": self.heartbeat,
            "stop_time": self.stop_time,
            "queue_time": self.queue_time,
            "status": self.status,
            "result": self.result,
            "meta": {"comment": self.comment, "priority": self.priority},
            "resources": [r.to_json() for r in self.resources],
            "artifacts": [a.to_json() for a in self.artifacts],
            "host": self.host.to_json(),
            "experiment": self.experiment.to_json(),
            "config": restore(json.loads(self.config)),
            "captured_out": self.captured_out,
            "fail_trace": self.fail_trace,
        }
